/* DC "Lan Adapter" example code

   (c)2002 Dan Potter (NEW BSD LICENSE)

   Happy New Year, everyone! =D Here is an example program which will
   (hopefully) activate your "DC Lan Adapter" (HIT-0300) and respond to any
   ping message pointed at it. It prints its MAC address when it starts up,
   so you can add an arp entry as usual.

   The DC Lan Adapter is an 8-bit G2 peripheral with the following block
   diagram:

   G2 <---> Altera PLD <---> MB86967 Ethernet <---> TDK "Pulse Transformer"
                                    ^                         ^
                                    |                         |
                                    v                         v
                              M5M5278DVP RAM            10BT ethernet

   Its IO base is located at 0xa060xxxx, like the modem, and the register
   interface seems very similar (I suspect they actually use the exact same
   PLD in both, to save money).

   As for the Fujitsu ethernet chip, the latest updated driver in BSD was
   for FreeBSD 2.x (!!) and was for PCMCIA, so a bit of interpolation was
   required. Fortunately the Fujitsu chip is _much_ simpler and easier to
   work with than the later RTL chip, but unfortunately, that simplicity
   comes at a cost. Each packet to be send or received must be processed
   over a PIO interface one byte at a time. This is probably why Sega dumped
   it for the RTL design later on, because you can imagine the system load
   to do that while trying to process 3D graphics and such...

   This driver is really simplistic, but this should form the basis with 
   which a better one can be written if neccessary. It was developed using
   KOS CVS (1.1.5+).

   If anyone has a DC Lan Adapter which this doesn't work with, please let
   me know!

   Thanks to AndrewK for the structs and the logic of the ping code.
   Thanks to TheGypsy for test hardware.

 */

#include <kos.h>

/* IP packet header structs (from dcload) */
typedef struct {
  unsigned char dest[6];
  unsigned char src[6];
  unsigned char type[2];
} ether_header_t;
#define ETHER_H_LEN 14

typedef struct {
  unsigned char version_ihl __attribute__ ((packed));
  unsigned char tos __attribute__ ((packed));
  unsigned short length __attribute__ ((packed));
  unsigned short packet_id __attribute__ ((packed));
  unsigned short flags_frag_offset __attribute__ ((packed));
  unsigned char ttl __attribute__ ((packed));
  unsigned char protocol __attribute__ ((packed));
  unsigned short checksum __attribute__ ((packed));
  unsigned int src __attribute__ ((packed));
  unsigned int dest __attribute__ ((packed));
} ip_header_t;

typedef struct {
  unsigned char type __attribute__ ((packed));
  unsigned char code __attribute__ ((packed));
  unsigned short checksum __attribute__ ((packed));
  unsigned int misc __attribute__ ((packed));
} icmp_header_t;

/* IP checksum (also from dcload) */
unsigned short checksum(unsigned short *buf, int count)
{
    unsigned long sum = 0;
                          
    while (count--) {     
        sum += *buf++;
        if (sum & 0xffff0000) {
            sum &= 0xffff;
            sum++;
        }
    }    
     return ~(sum & 0xffff);
}


/* Some basic macros to assist with register access */
#define REGL(a) (vuint32 *)(a)
#define REGS(a) (vuint16 *)(a)
#define REGC(a) (vuint8 *)(a)

vuint8 *xpc = REGC(0xa0600000);
vuint16 *xps = REGS(0xa0600000);
vuint32 *xpl = REGL(0xa0600000);
#define REG(x) ( xpl[(x) + 0x400/4] & 0xff )
#define REGW(x) ( xpc[(x)*4 + 0x400] )
	
/* This is based on the JLI EEPROM reader from FreeBSD. EEPROM in the
   Sega adapter is a bit simpler than what is described in the Fujitsu
   manual -- it appears to contain only the MAC address and not a base
   address like the manual says. EEPROM is read one bit (!) at a time
   through the EEPROM interface port. */
#define FE_B16_SELECT	0x20		/* EEPROM chip select */
#define FE_B16_CLOCK	0x40		/* EEPROM shift clock */
#define FE_B17_DATA	0x80		/* EEPROM data bit */
void net_strobe_eeprom() {
	REGW(16) = FE_B16_SELECT;
	REGW(16) = FE_B16_SELECT | FE_B16_CLOCK;
	REGW(16) = FE_B16_SELECT | FE_B16_CLOCK;
	REGW(16) = FE_B16_SELECT;
}
void net_read_eeprom(uint8 *data) {
	uint8 save16, save17, val;
	int n, bit;

	/* Save the current value of the EEPROM registers */
	save16 = REG(16);
	save17 = REG(17);

	/* Read bytes from EEPROM, two per iteration */
	for (n=0; n<3; n++) {
		/* Reset the EEPROM interface */
		REGW(16) = 0;
		REGW(17) = 0;

		/* Start EEPROM access */
		REGW(16) = FE_B16_SELECT;
		REGW(17) = FE_B17_DATA;
		net_strobe_eeprom();

		/* Pass the iteration count as well as a READ command */
		val = 0x80 | n;
		for (bit=0x80; bit != 0x00; bit>>=1) {
			REGW(17) = (val & bit) ? FE_B17_DATA : 0;
			net_strobe_eeprom();
		}
		REGW(17) = 0;

		/* Read a byte */
		val = 0;
		for (bit=0x80; bit != 0x00; bit>>=1) {
			net_strobe_eeprom();
			if (REG(17) & FE_B17_DATA)
				val |= bit;
		}
		*data++ = val;

		/* Read one more byte */
		val = 0;
		for (bit=0x80; bit != 0x00; bit>>=1) {
			net_strobe_eeprom();
			if (REG(17) & FE_B17_DATA)
				val |= bit;
		}
		*data++ = val;
	}

	printf("  read 6 bytes of EEPROM data:");
	data -= 6;
	for (n=0; n<6; n++)
		printf(" %02X", data[n]);
	printf("\n");
}

/* To select a register bank */
#define SETBANK(x) do { \
	int i = (REG(7) & ~0x0c) | ((x) << 2); \
	REGW(7) = i; \
} while(0)

/* Reset the lan adapter and set it up for send/receive */
void net_reset() {
	int i, type;
	uint8 mac[6];
	
	/* Reset the interface */
	printf("resetting expansion interface...\n");
	xpc[0x0480] = 0;
	xpc[0x0480] = 1;
	xpc[0x0480] = 0;	/* Leave it in write-enabled mode */

	/* Give it a few MS to come back */
	timer_spin_sleep(100);

	/* Clear interrupt status */
	REGW(0) = 0xff;
	REGW(1) = 0xff;

	/* Power down chip */
	timer_spin_sleep(2);
	REGW(7) = REG(7) & ~0x20;
	timer_spin_sleep(2);

	/* Disable rx/tx */
	timer_spin_sleep(2);
	REGW(6) = REG(6) | 0x80;
	timer_spin_sleep(2);

	/* Power up the chip */
	timer_spin_sleep(2);
	REGW(7) = REG(7) | 0x20;
	timer_spin_sleep(2);

	/* Read the chip type */
	printf("  chip ID is ");
	type = (REG(7) >> 6) & 3;
	switch (type) {
	case 0:
		printf("MB86960A");
		break;
	case 1:
		printf("MB86964");
		break;
	case 2:
		printf("MB86967");
		break;
	case 3:
		printf("MB86965A");
		break;
	}
	printf("\n");
	if (type != 2) {
		printf("    chip type is not correct, this probly won't work\n");
	}

	/* Read the EEPROM data */
	net_read_eeprom(mac);

	/* Select BANK0 and write the MAC address */
	SETBANK(0);
	timer_spin_sleep(2);
	for (i=0; i<6; i++) {
		REGW(i+8) = mac[i];
	}
	
	/* Read out the MAC address (to test) */
	printf("  mac address: ");
	for (i=0; i<6; i++) {
		if (i < 5)
			printf("%02x:", (int)REG(i+8));
		else
			printf("%02x\n", (int)REG(i+8));
	}

	/* Clear the multicast address */
	SETBANK(1);
	for (i=0; i<6; i++)
		REGW(i+8) = 0;

	/* Select the BMPR bank for normal operation */
	SETBANK(2);
	REGW(10) = 0;
	REGW(11) = 0x01 | 0x02;		/* Reset any collision/skip counters */
	REGW(12) = 0;
	REGW(13) = 0;
	REGW(14) = 0;
	REGW(15) = 0;

	/* Enable transmitter / receiver */
	REGW(6) = REG(6) & ~0x80;

	/* Set non-promiscuous mode (use 0x03 for promiscuous) */
	REGW(5) = (REG(5) | 0x02) & ~0x01;
}

/* Shutdown lan adapter */
void net_stop() {
	/* Disable transmitter / receiver */
	REGW(6) = REG(6) | 0x80;

	/* Power down */
	timer_spin_sleep(2);
	REGW(7) = REG(7) & ~0x20;
	timer_spin_sleep(2);
}

/* Transmit a packet */
/* Note that it's technically possible to queue up more than one packet 
   at a time for transmission, but this is the simple way. */
int net_tx(char *pkt, int len) {
	int i;

	/* XXX check for overflow/pending transmit */

	/* Poke the length */
	REGW(8) = (len & 0x00ff);
	REGW(8) = (len & 0xff00) >> 8;

	/* Write the packet */
	for (i=0; i<len; i++)
		REGW(8) = pkt[i];

	/* Start the transmitter */
	REGW(10) = 1 | 0x80;	/* 1 packet, 0x80 = start */

	return 1;
}

/* Packet buffer */
uint8 packet[1514];
ether_header_t *ether_header = (ether_header_t *)(packet + 0);
ip_header_t *ip_header = (ip_header_t *)(packet + 14);

int ntohs(int);

/* Process a received ping */
void process_ping() {
	icmp_header_t *icmp;
	uint8 tmp[6];

	/* Find the header */
	icmp = (icmp_header_t *)(packet + ETHER_H_LEN + 4*(ip_header->version_ihl & 0x0f));

	if (icmp->type != 8) {
		printf("Unknown ICMP type %d received\n", icmp->type);
		return;
	}

	/* pong */
	icmp->type = 0;

	/* Swap src and dest hardware addresses */
	memcpy(tmp, ether_header->dest, 6);
	memcpy(ether_header->dest, ether_header->src, 6);
	memcpy(ether_header->src, tmp, 6);

	/* Swap src and dest IP addresses */
	memcpy(tmp, &ip_header->src, 4);
	memcpy(&ip_header->src, &ip_header->dest, 4);
	memcpy(&ip_header->dest, tmp, 4);

	/* Recompute IP checksum */
	ip_header->checksum = 0;
	ip_header->checksum = checksum((unsigned short *)ip_header, 2*(ip_header->version_ihl & 0x0f));

	/* Recompute ICMP checksum */
	icmp->checksum = 0;
	icmp->checksum = checksum((unsigned short *)icmp, ntohs(ip_header->length)/2 - 2*(ip_header->version_ihl & 0x0f));

	/* Transmit it */
	net_tx(packet, ETHER_H_LEN + ntohs(ip_header->length));
}
	
/* Check for received packets */
int net_recv() {
	int i, status, len, count;

	for (count = 0; ; count++) {
		/* Is the buffer empty? */
		if (REG(5) & 0x40)
			return count;

		/* Get the receive status byte */
		status = REG(8);
		(void)REG(8);

		/* Get the packet length */
		len = REG(8);
		len |= REG(8) << 8;

		/* Check for errors */
		if ( (status & 0xF0) != 0x20 ) {
			printf("Received error code: %02x, len is %04x\n", status, len);
			return -1;
		}

		/* Read the packet */
		printf("Received packet of size %d:\n", len);
		if (len > 1514) {
			printf("Packet was too large, aborting\n");
			return -2;
		}
		for (i=0; i<len; i++) {
			packet[i] = REG(8);
		}

		/* Write it to the console; uncomment this if you wanna see */
		/* for (i=0; i<len; i++) {
			if (i && !(i % 16))
				printf("\n");
			printf("%02x ", packet[i]);
		}
		printf("\n\n"); */

		/* If it's a ping, then send a pong */
		if (ip_header->protocol == 1) {
			process_ping();
		} else {
			printf("Received a packet, but it wasn't a ping\n");
		}
	}
}

/* Loop doing something interesting */
void net_loop() {
	cont_cond_t cond;
	int cont;

	printf("Entering receive loop.. press START on the first controller to quit!\n");

	cont = maple_first_controller();
	for (;;) {
		/* Check for a controller */
		if (cont_get_cond(cont, &cond) < 0) {
			printf("Can't read controller\n");
			return;
		}
		if (!(cond.buttons & CONT_START))
			return;

		/* Check for received packets */
		if (net_recv() < 0)
			return;

		/* Delay to make sure the controller doesn't freak out */
		timer_spin_sleep(1000/60);
	}
}

int main(int argc, char **argv) {
	kos_init_all(IRQ_ENABLE, NULL);
	net_reset();
	net_loop();
	net_stop();
}

